
"""
Import winpty.h struct and function definitions.
"""

cimport cython

# Import some windows.h type shorthands
cdef extern from "Windows.h":
    ctypedef Py_UNICODE WCHAR
    ctypedef const WCHAR* LPCWSTR
    ctypedef void* HWND
    ctypedef void* HANDLE
    ctypedef unsigned long long UINT64
    ctypedef unsigned long DWORD


# Import winpty.h C++ definitions
cdef extern from 'winpty.h':
    # Error constants and structs
    ctypedef struct winpty_error_t
    ctypedef winpty_error_t* winpty_error_ptr_t
    ctypedef DWORD winpty_result_t

    # Error related functions
    winpty_result_t winpty_error_code(winpty_error_ptr_t err)
    LPCWSTR winpty_error_msg(winpty_error_ptr_t err)
    void winpty_error_free(winpty_error_ptr_t err)

    # Winpty agent configuration settings
    # ctypedef struct winpty_config_s winpty_config_t
    ctypedef struct winpty_config_t

    # Winpty agent configuration functions
    winpty_config_t* winpty_config_new(UINT64 agentFlags, winpty_error_ptr_t *err)
    void winpty_config_free(winpty_config_t *cfg)
    void winpty_config_set_initial_size(winpty_config_t *cfg, int cols, int rows)
    void winpty_config_set_mouse_mode(winpty_config_t *cfg, int mouseMode)
    void winpty_config_set_agent_timeout(winpty_config_t *cfg, DWORD timeoutMs)

    # Agent start related structs
    # ctypedef struct winpty_s winpty_t
    ctypedef struct winpty_t

    # Start a new agent
    winpty_t * winpty_open(const winpty_config_t *cfg, winpty_error_ptr_t *err)
    HANDLE winpty_agent_process(winpty_t *wp)

    # I/O Pipes
    LPCWSTR winpty_conin_name(winpty_t *wp)
    LPCWSTR winpty_conout_name(winpty_t *wp)
    LPCWSTR winpty_conerr_name(winpty_t *wp)


    # Process creation configuration struct
    ctypedef struct winpty_spawn_config_t

    # Process creation functions
    winpty_spawn_config_t * winpty_spawn_config_new(UINT64 spawnFlags,
                                                    LPCWSTR appname,
                                                    LPCWSTR cmdline,
                                                    LPCWSTR cwd,
                                                    LPCWSTR env,
                                                    winpty_error_ptr_t *err)

    void winpty_spawn_config_free(winpty_spawn_config_t *cfg)

    # Spawn new processes
    bint winpty_spawn(winpty_t *wp,
                      const winpty_spawn_config_t *cfg,
                      HANDLE *process_handle,
                      HANDLE *thread_handle,
                      DWORD *create_process_error,
                      winpty_error_ptr_t *err)

    # Winpty agent RPC calls
    bint winpty_set_size(winpty_t *wp, int cols, int rows,
                         winpty_error_ptr_t *err)

    # Free winpty_t pointer
    void winpty_free(winpty_t *wp)




"""
Cython wrapper around Winpty and Windows functions.
"""

cimport cython
from winpty._winpty cimport winpty, winpty_constants

# Windows API types
cdef extern from "Windows.h":
    ctypedef Py_UNICODE WCHAR
    ctypedef const WCHAR* LPCWSTR
    ctypedef void* HWND
    ctypedef void* HANDLE
    ctypedef unsigned long long UINT64
    ctypedef unsigned long DWORD
    ctypedef DWORD *LPDWORD
    ctypedef LPCWSTR LPCTSTR

    DEF STILL_ACTIVE = 259

    DWORD GetProcessId(HANDLE proc)
    bint GetExitCodeProcess(
      HANDLE  hProcess,
      LPDWORD lpExitCode
    )


cdef class Agent:
    """
    This class wraps a winpty agent and offers communication with
    a winpty-agent process.
    """
    cdef winpty.winpty_t* _c_winpty_t
    cdef HANDLE _agent_process
    cdef HANDLE _process
    cdef DWORD pid
    cdef LPCWSTR conin_pipe_name
    cdef LPCWSTR conout_pipe_name
    cdef DWORD exitstatus
    cdef unsigned char alive

    def __init__(self, int cols, int rows, bint override_pipes=True,
                 int mouse_mode=winpty_constants._WINPTY_MOUSE_MODE_NONE,
                 int timeout=30000, int agent_config=winpty_constants._WINPTY_FLAG_COLOR_ESCAPES):
        """
        Initialize a winpty agent wrapper of size ``(cols, rows)``.
        """
        cdef winpty.winpty_error_ptr_t err
        cdef winpty.winpty_config_t* config = winpty.winpty_config_new(agent_config, &err)

        if config is NULL:
            msg = 'An error has ocurred: {0} - Code: {1}'.format(
                winpty.winpty_error_msg(err),
                winpty.winpty_error_code(err))
            winpty.winpty_error_free(err)
            raise RuntimeError(msg)

        if cols <= 0 or rows <= 0:
            msg = 'PTY cols and rows must be positive and non-zero. Got: ({0}, {1})'.format(
                cols, rows)
            raise RuntimeError(msg)

        winpty.winpty_config_set_initial_size(config, cols, rows)
        winpty.winpty_config_set_mouse_mode(config, mouse_mode)
        winpty.winpty_config_set_agent_timeout(config, timeout)

        cdef winpty.winpty_error_ptr_t err_pointer
        self._c_winpty_t = winpty.winpty_open(config, &err_pointer)
        winpty.winpty_config_free(config)

        if self._c_winpty_t is NULL:
            msg = 'An error has ocurred: {0} - Code: {1}'.format(
                winpty.winpty_error_msg(err_pointer),
                winpty.winpty_error_code(err_pointer))
            winpty.winpty_error_free(err_pointer)
            raise RuntimeError(msg)

        self._agent_process = winpty.winpty_agent_process(self._c_winpty_t)
        self.conin_pipe_name = winpty.winpty_conin_name(self._c_winpty_t)
        self.conout_pipe_name = winpty.winpty_conout_name(self._c_winpty_t)
        self.alive = 0
        self.pid = 0

    property conin_pipe_name:
        def __get__(self):
            return self.conin_pipe_name

    property conout_pipe_name:
        def __get__(self):
            return self.conout_pipe_name

    property pid:
        def __get__(self):
            return self.pid

    property exitstatus:
        def __get__(self):
            if self.pid == 0:
                return None
            if self.alive == 1:
                self.isalive()
            if self.alive == 1:
                return None
            return self.exitstatus

    def spawn(self, LPCWSTR appname, LPCWSTR cmdline=NULL,
              LPCWSTR cwd=NULL, LPCWSTR env=NULL):
        """
        Start a windows process that communicates through a winpty-agent process.
        """
        cdef winpty.winpty_error_ptr_t spawn_conf_err
        cdef winpty.winpty_spawn_config_t* spawn_config = winpty.winpty_spawn_config_new(winpty_constants._WINPTY_SPAWN_FLAG_MASK,
                                                                                         appname, cmdline, cwd, env, &spawn_conf_err)
        if spawn_config is NULL:
            msg = 'An error has ocurred: {0} - Code: {1}'.format(
                winpty.winpty_error_msg(spawn_conf_err),
                winpty.winpty_error_code(spawn_conf_err))
            winpty.winpty_error_free(spawn_conf_err)
            raise RuntimeError(msg)

        cdef winpty.winpty_error_ptr_t spawn_err
        cdef bint succ = winpty.winpty_spawn(self._c_winpty_t, spawn_config, &self._process,
                                             NULL, NULL, &spawn_err)
        winpty.winpty_spawn_config_free(spawn_config)

        if not succ:
            msg = 'An error has ocurred: {0} - Code: {1}'.format(
                winpty.winpty_error_msg(spawn_err),
                winpty.winpty_error_code(spawn_err))
            winpty.winpty_error_free(spawn_err)
            raise RuntimeError(msg)

        self.pid = GetProcessId(self._process)
        self.alive = 1

        return succ

    def set_size(self, int cols, int rows):
        """
        Resize the console size of current agent process.
        """
        if cols <= 0 or rows <= 0:
            msg = 'PTY cols and rows must be positive and non-zero. Got: ({0}, {1})'.format(
                cols, rows)
            raise RuntimeError(msg)

        cdef winpty.winpty_error_ptr_t err_pointer
        cdef bint succ = winpty.winpty_set_size(self._c_winpty_t, cols, rows, &err_pointer)

        if not succ:
            msg = 'An error has ocurred: {0} - Code: {1}'.format(
                winpty.winpty_error_msg(err_pointer),
                winpty.winpty_error_code(err_pointer))
            winpty.winpty_error_free(err_pointer)
            raise RuntimeError(msg)

    def isalive(self):
        """This tests if the pty process is running or not.
        This is non-blocking.
        """
        cdef DWORD lpExitCode
        cdef bint succ = GetExitCodeProcess(self._process, &lpExitCode)
        if not succ:
            raise RuntimeError('Could not check status')

        # Check for STILL_ACTIVE flag
        # https://msdn.microsoft.com/en-us/library/windows/desktop/ms683189(v=vs.85).aspx
        alive = lpExitCode == STILL_ACTIVE
        if not alive:
            self.alive = 0
            self.exitstatus = lpExitCode
        return alive

    def __dealloc__(self):
        if self._c_winpty_t is not NULL:
            winpty.winpty_free(self._c_winpty_t)



# -*- coding: utf-8 -*-
"""winpty wrapper tests."""

# yapf: disable

# Standard library imports
import os

# Third party imports
from flaky import flaky
from winpty.ptyprocess import which
from winpty.winpty_wrapper import PTY, PY2
import pytest


# yapf: enable

CMD = which('cmd')
if PY2:
    CMD = unicode(CMD)  # noqa


@pytest.fixture(scope='module')
def pty_fixture(cols, rows):
    pty = PTY(cols, rows)
    pty.spawn(CMD)
    return pty


@flaky(max_runs=4, min_passes=1)
def test_read():
    pty = pty_fixture(80, 25)
    loc = os.getcwd()
    line = ''
    while loc not in line:
        line += pty.read().decode('utf-8')
    assert loc in line
    pty.close()
    del pty


def test_write():
    pty = pty_fixture(80, 25)
    line = pty.read()
    while len(line) < 10:
        line = pty.read()

    text = u'Eggs, ham and spam Ã¼nicode'
    pty.write(text)

    line = u''
    while text not in line:
        line += pty.read().decode('utf-8')

    assert text in line

    pty.close()
    del pty


def test_isalive():
    pty = pty_fixture(80, 25)
    pty.write(u'exit\r\n')

    text = u'exit'
    line = u''
    while text not in line:
        line += pty.read().decode('utf-8')

    while pty.isalive():
        pty.read()
        continue

    assert not pty.isalive()
    pty.close()
    del pty




"""
Winpty header constants.
"""

cdef extern from 'winpty_constants.h':
    # Agent error constants
    cdef int _WINPTY_ERROR_SUCCESS "WINPTY_ERROR_SUCCESS"
    cdef int _WINPTY_ERROR_OUT_OF_MEMORY "WINPTY_ERROR_OUT_OF_MEMORY"
    cdef int _WINPTY_ERROR_SPAWN_CREATE_PROCESS_FAILED "WINPTY_ERROR_SPAWN_CREATE_PROCESS_FAILED"
    cdef int _WINPTY_ERROR_LOST_CONNECTION "WINPTY_ERROR_LOST_CONNECTION"
    cdef int _WINPTY_ERROR_AGENT_EXE_MISSING "WINPTY_ERROR_AGENT_EXE_MISSING"
    cdef int _WINPTY_ERROR_UNSPECIFIED "WINPTY_ERROR_UNSPECIFIED"
    cdef int _WINPTY_ERROR_AGENT_DIED "WINPTY_ERROR_AGENT_DIED"
    cdef int _WINPTY_ERROR_AGENT_TIMEOUT "WINPTY_ERROR_AGENT_TIMEOUT"
    cdef int _WINPTY_ERROR_AGENT_CREATION_FAILED "WINPTY_ERROR_AGENT_CREATION_FAILED"

    # Agent configuration constants
    cdef int _WINPTY_FLAG_CONERR "WINPTY_FLAG_CONERR"
    cdef int _WINPTY_FLAG_PLAIN_OUTPUT "WINPTY_FLAG_PLAIN_OUTPUT"
    cdef int _WINPTY_FLAG_COLOR_ESCAPES "WINPTY_FLAG_COLOR_ESCAPES"
    cdef int _WINPTY_FLAG_ALLOW_CURPROC_DESKTOP_CREATION "WINPTY_FLAG_ALLOW_CURPROC_DESKTOP_CREATION"

    # All agent configuration constants enabled
    cdef int _WINPTY_FLAG_MASK "WINPTY_FLAG_MASK"

    # Disable/Enable mouse input capture
    cdef int _WINPTY_MOUSE_MODE_NONE "WINPTY_MOUSE_MODE_NONE"
    cdef int _WINPTY_MOUSE_MODE_AUTO "WINPTY_MOUSE_MODE_AUTO"
    cdef int _WINPTY_MOUSE_MODE_FORCE "WINPTY_MOUSE_MODE_FORCE"

    # General winpty-agent process behaviour constants
    cdef int _WINPTY_SPAWN_FLAG_AUTO_SHUTDOWN "WINPTY_SPAWN_FLAG_AUTO_SHUTDOWN"
    cdef int _WINPTY_SPAWN_FLAG_EXIT_AFTER_SHUTDOWN "WINPTY_SPAWN_FLAG_EXIT_AFTER_SHUTDOWN"
    cdef int _WINPTY_SPAWN_FLAG_MASK "WINPTY_SPAWN_FLAG_MASK"



# -*- coding: utf-8 -*-
"""Setup script for pywinpty."""

# yapf: disable

# Standard library imports
import ast
import os

# Third party imports
from setuptools import Extension, find_packages, setup


# yapf: enable

HERE = os.path.abspath(os.path.dirname(__file__))


def get_version(module='winpty'):
    """Get version from text file and avoids importing the module."""
    with open(os.path.join(HERE, module, '__init__.py'), 'r') as f:
        data = f.read()
    lines = data.split('\n')
    for line in lines:
        if line.startswith('VERSION_INFO'):
            version_tuple = ast.literal_eval(line.split('=')[-1].strip())
            version = '.'.join(map(str, version_tuple))
            break
    return version


try:
    include_dirs = [os.environ['LIBRARY_INC']]
except KeyError:
    include_dirs = []
try:
    library_dirs = [os.environ['LIBRARY_LIB']]
except KeyError:
    library_dirs = []


ext_options = {}
cythonize_options = {}
if os.environ.get('CYTHON_COVERAGE'):
    cythonize_options['compiler_directives'] = {'linetrace': True}
    cythonize_options['annotate'] = True
    ext_options['define_macros'] = [
        ('CYTHON_TRACE', '1'), ('CYTHON_TRACE_NOGIL', '1')
    ]

try:
    from Cython.Build import cythonize
    ext_modules = cythonize([
        Extension(
            "winpty.cywinpty",
            sources=["winpty/cywinpty.pyx"],
            libraries=["winpty"],
            include_dirs=include_dirs,
            library_dirs=library_dirs,
            **ext_options
        )],
        **cythonize_options
    )
except ImportError:
    ext_modules = []


setup(
    name='pywinpty',
    version=get_version(),
    keywords=['winpty', 'pty', 'pseudoterminal', 'pseudotty'],
    url='https://github.com/spyder-ide/pywinpty',
    license='MIT',
    author='Edgar Andrés Margffoy-Tuay',
    author_email='andfoy@gmail.com',
    description='Python bindings for the winpty library',
    ext_modules=ext_modules,
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    setup_requires=['Cython'],
    package_data=dict(winpty=['*.pyd', '*.dll', '*.exe']),
    install_requires=['backports.shutil_which;python_version<"3.0"'],
    classifiers=[
        'Development Status :: 4 - Beta', 'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ]
)
